#!/usr/bin/python

import errno
import os
import Queue
import signal
import sys
import threading
import time
import unittest

from bzrlib import (
    tests,
    transform,
    )


try:
    from debian import changelog
except ImportError:
    from debian_bundle import changelog

sys.path.insert(0, os.path.dirname(__file__))
import icommon
import import_package


class PackageToImportTests(unittest.TestCase):

    def test_same_as_same(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="main",
                on_lp=True, url=None)
        b = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="contrib")
        c = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", on_lp=False)
        d = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", url="a")
        self.assertTrue(a.same_as(b))
        self.assertTrue(a.same_as(c))
        self.assertTrue(a.same_as(d))

    def test_same_as_different(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="main",
                on_lp=True, url=None)
        b = icommon.PackageToImport("b", changelog.Version("1"),
                "debian", "sid", "release")
        c = icommon.PackageToImport("a", changelog.Version("1.1"),
                "debian", "sid", "release")
        d = icommon.PackageToImport("a", changelog.Version("1"),
                "ubuntu", "sid", "release")
        e = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "squeeze", "release")
        f = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "security")
        self.assertFalse(a.same_as(b))
        self.assertFalse(a.same_as(c))
        self.assertFalse(a.same_as(d))
        self.assertFalse(a.same_as(e))
        self.assertFalse(a.same_as(f))

    def test_get_url(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", url="url")
        self.assertEqual("url", a.get_url())
        b = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="non-free", on_lp=False)
        self.assertEqual(icommon.urllib_debian_base_url
                + "pool/non-free/a/a/a_1.dsc", b.get_url())
        c = icommon.PackageToImport("libfoo", changelog.Version("1:1.1-1"),
                "debian", "squeeze", "release", component="contrib", on_lp=False)
        self.assertEqual(icommon.urllib_debian_base_url
                + "pool/contrib/libf/libfoo/libfoo_1.1-1.dsc", c.get_url())
        c = icommon.PackageToImport("libfoo", changelog.Version("1:1.1-1"),
                "debian", "squeeze", "release", component="contrib",
                on_lp=True)
        self.assertEqual(icommon.urllib_launchpad_base_url
                + "debian/squeeze/+source/libfoo/1:1.1-1/+files/libfoo_1.1-1.dsc",
                c.get_url())
        d = icommon.PackageToImport("libfoo", changelog.Version("1:1.1-1"),
                "ubuntu", "lucid", "security")
        self.assertEqual(icommon.urllib_launchpad_base_url
                + "ubuntu/lucid/+source/libfoo/1:1.1-1/+files/libfoo_1.1-1.dsc",
                d.get_url())

    def test_str(self):
        a = icommon.PackageToImport("a", changelog.Version("1"),
                "debian", "sid", "release", component="main",
                on_lp=True, url=None)
        self.assertEqual("PackageToImport(a, 1, debian, sid, release, main)",
                str(a))


class ImportListTests(unittest.TestCase):

    def make_simple(self):
        self.pkg1 = icommon.PackageToImport("pkg", "0.1-1", "ubuntu",
                "jaunty", "Release")
        self.pkg2 = icommon.PackageToImport("pkg", "1.0-1", "ubuntu",
                "karmic", "Release")
        list = icommon.ImportList()
        list.add_if_needed(self.pkg1)
        list.add_if_needed(self.pkg2)
        return list

    def test_str(self):
        list = self.make_simple()
        self.assertEqual("[PackageToImport(pkg, 0.1-1, ubuntu, "
                "jaunty, Release, ), PackageToImport(pkg, 1.0-1, "
                "ubuntu, karmic, Release, )]", str(list))

    def test_reverse(self):
        list = self.make_simple()
        self.assertEqual([self.pkg1, self.pkg2], list.plist)
        list.reverse()
        self.assertEqual([self.pkg2, self.pkg1], list.plist)

    def test_sort(self):
        list = self.make_simple()
        self.assertEqual([self.pkg1, self.pkg2], list.plist)
        list.sort()
        self.assertEqual([self.pkg2, self.pkg1], list.plist)

    def test_add_if_needed(self):
        list = self.make_simple()
        self.assertEqual([self.pkg1, self.pkg2], list.plist)
        list.add_if_needed(self.pkg1)
        self.assertEqual([self.pkg1, self.pkg2], list.plist)


class RevidDatabaseTests(tests.TestCaseWithTransport):

    def setUp(self):
        super(RevidDatabaseTests, self).setUp()
        self.cleanup = 0
        self.push = 0
        self.db = self.get_db()
        self.version = changelog.Version("1.0-1")
        self.version2 = changelog.Version("1.0-2")
        self.version3 = changelog.Version("1.0-3")
        self.suite = "karmic"
        self.tree = self.make_branch_and_tree(".")
        self.revid = self.tree.commit("one")
        self.revid2 = self.tree.commit("two")
        self.revid3 = self.tree.commit("three")

    def get_db(self):
        return icommon.RevidDatabase(":memory:", "pkg")

    def check_rows(self, memory, working, saved):
        self.assertEqual(memory, len(self.db._memory_rows()))
        self.assertEqual(working, len(self.db._working_rows()))
        self.assertEqual(saved, len(self.db._saved_rows()))

    def test_mark(self):
        self.assertFalse(self.db.is_marked(self.version, self.suite))
        self.check_rows(0, 0, 0)
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.check_rows(1, 0, 0)
        self.assertTrue(self.db.is_marked(self.version, self.suite))

    def test_mark_again(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.assertRaises(AssertionError, self.db.commit_outstanding)
        self.assertRaises(AssertionError, self.db.is_marked, self.version,
                self.suite)

    def test_check(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.assertTrue(self.db.check(self.version, self.revid, self.suite,
                    self.tree.branch))
        self.assertFalse(self.db.check(changelog.Version(
                        str(self.version)+"1"),
                    self.revid, self.suite, self.tree.branch))
        self.assertFalse(self.db.check(self.version, self.revid,
                    self.suite+"a", self.tree.branch))

    def test_commit(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.check_rows(0, 0, 1)
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.assertRaises(AssertionError, self.db.commit_outstanding)
        self.check_rows(1, 0, 1)

    def do_cleanup(self):
        self.cleanup += 1

    def do_push(self):
        self.push += 1

    def test_cleanup_nothing_to_do(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit()
        self.db.cleanup_last_run(self.do_cleanup, self.do_push)
        self.assertEqual(0, self.push)
        self.assertEqual(1, self.cleanup)

    def test_cleanup(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.db.mark(self.version2, self.revid2, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.check_rows(0, 1, 1)
        self.db.cleanup_last_run(self.do_cleanup, self.do_push)
        self.assertEqual(1, self.push)
        self.assertEqual(1, self.cleanup)
        self.check_rows(0, 0, 2)

    def test_discard(self):
        self.db.mark(self.version, self.revid, self.suite, self.tree.branch)
        self.db.mark(self.version2, self.revid2, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.db.commit()
        self.db.mark(self.version3, self.revid3, self.suite,
                self.tree.branch)
        self.db.commit_outstanding()
        self.check_rows(0, 1, 2)
        self.db.discard_last_run()
        self.check_rows(0, 0, 2)


class FindEarliestMergeTests(tests.TestCaseWithTransport):

    def test_tip_revision(self):
        t = self.make_branch_and_tree(".")
        revid = t.commit("one")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, revid))

    def test_a_parent(self):
        t = self.make_branch_and_tree(".")
        revid = t.commit("one")
        t.commit("two")
        t.commit("three")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, revid))

    def test_merged(self):
        t = self.make_branch_and_tree(".")
        t.commit("one")
        t2 = t.bzrdir.sprout("merged").open_workingtree()
        merged_revid = t2.commit("one and a half")
        t.merge_from_branch(t2.branch)
        revid = t.commit("two")
        t.commit("three")
        self.assertEquals(revid,
                icommon.find_earliest_merge(t.branch, merged_revid))


class TestCheckSame(tests.TestCaseWithTransport):

    def assertTreesEquivalent(self, repo, rev1, rev2):
        rtree1 = repo.revision_tree(rev1)
        rtree2 = repo.revision_tree(rev2)
        self.assertTrue(
            import_package._are_trees_equivalent(rtree1, rtree2),
            'Trees are not actually equivalent')

    def assertTreesNotEquivalent(self, repo, rev1, rev2):
        rtree1 = repo.revision_tree(rev1)
        rtree2 = repo.revision_tree(rev2)
        self.assertFalse(
            import_package._are_trees_equivalent(rtree1, rtree2),
            'Trees are accidentally equivalent')

    def test_identical_different_revs_same_fids(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/file', 'content\n')])
        t1.add(['file'], ['file-id'])
        revid = t1.commit('first')
        t2 = t1.bzrdir.sprout('tree2').open_workingtree()
        # Both trees independently get identical content
        self.build_tree_contents([('tree1/file', 'new content\n'),
                                  ('tree2/file', 'new content\n')])
        t1_rev = t1.commit('new content')
        t2_rev = t2.commit('new content')
        t1.branch.repository.fetch(t2.branch.repository, t2_rev)

        self.assertTreesEquivalent(repo, t1_rev, t2_rev)

    def test_exec_patch_change_is_ignored(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/patches/',),
                                  ('tree1/debian/patches/my.patch', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/patches', 'debian/patches/my.patch'])
        rev1 = t1.commit('create the new tree')
        tt = transform.TreeTransform(t1)
        tt.set_executability(True,
            tt.trans_id_file_id(t1.path2id('debian/patches/my.patch')))
        tt.apply()
        rev2 = t1.commit('set executable bit')
        self.assertTreesEquivalent(repo, rev1, rev2)
        self.build_tree_contents([
            ('tree1/debian/patches/my.patch', 'new content\n'),])
        rev3 = t1.commit('change content')
        self.assertTreesNotEquivalent(repo, rev1, rev3)
        self.assertTreesNotEquivalent(repo, rev2, rev3)

    def test_deleting_content_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/file', 'content\n'),
                                  ('tree1/debian/dir/file2', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/file', 'debian/dir/file2'])
        rev1 = t1.commit('initial')
        os.remove('tree1/debian/dir/file2')
        rev2 = t1.commit('del file2')
        self.assertTreesNotEquivalent(repo, rev1, rev2)

    def test_deleting_last_content_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/file', 'content\n'),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/file'])
        rev1 = t1.commit('initial')
        os.remove('tree1/debian/dir/file')
        rev2 = t1.commit('del file2')
        self.assertTreesNotEquivalent(repo, rev1, rev2)

    def test_deleting_cascaded_dirs_is_not_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/debian/dir/subdir1/',),
                                  ('tree1/debian/dir/subdir2/',),
                                  ])
        t1.add(['debian', 'debian/dir', 'debian/dir/subdir1',
                'debian/dir/subdir2'])
        rev1 = t1.commit('initial')
        os.rmdir('tree1/debian/dir/subdir1')
        os.rmdir('tree1/debian/dir/subdir2')
        os.rmdir('tree1/debian/dir')
        rev2 = t1.commit('del subtree')
        self.assertTreesEquivalent(repo, rev1, rev2)
        # The reverse change is *not* equivalent
        self.assertTreesNotEquivalent(repo, rev2, rev1)

    def test_deleting_non_debian_empty_dir_is_a_change(self):
        t1 = self.make_branch_and_tree("tree1")
        repo = t1.branch.repository
        self.build_tree_contents([('tree1/debian/',),
                                  ('tree1/debian/dir/',),
                                  ('tree1/subdir1/',),
                                  ])
        t1.add(['debian', 'debian/dir', 'subdir1'])
        rev1 = t1.commit('initial')
        os.rmdir('tree1/subdir1')
        rev2 = t1.commit('del subbdir')
        self.assertTreesNotEquivalent(repo, rev1, rev2)


class ExceptionForTests(Exception):
    """This is intended to be used by tests to identify a specific exception.

    This allows test to catch this exception instead of a possibly ambiguous
    one.
    """
    pass


class TestSubprocessMonitor(tests.TestCase):

    def run_in_subprocess(self, args, klass=None):
        if klass is None:
            klass = icommon.SubprocessMonitor
        sub = klass(args)
        self.addCleanup(sub.join, 0)
        sub.start()
        return sub

    def run_sleeper_in_subprocess(self, klass=None):
        """Run a sleeper in a subprocess.

        This is used to simulate a never-ending process without consuming too
        much CPU cycles.
        """
        cmd = 'import time; while True: time.sleep(0.001)'
        return self.run_in_subprocess(['-c', cmd], klass=klass)

    def test_broken_command(self):
        sub = self.run_in_subprocess(['this-is-not-a-python-script'])
        sub.stopped.wait()
        self.assertEquals(True, sub.started.isSet())
        self.assertEquals(
            "python: can't open file 'this-is-not-a-python-script':"
            " [Errno 2] No such file or directory\n",
            sub.err)

    def test_success(self):
        sub = self.run_in_subprocess(['-c', 'print "toto"'])
        sub.stopped.wait()
        self.assertEquals('toto\n', sub.out)
        self.assertEquals('', sub.err)
        self.assertEquals(0, sub.retcode)

    def test_failure(self):
        sub = self.run_in_subprocess(['-c', 'import sys; sys.exit(1)'])
        sub.stopped.wait()
        self.assertEquals('', sub.out)
        self.assertEquals('', sub.err)
        self.assertEquals(1, sub.retcode)

    def test_kill(self):
        sub = self.run_sleeper_in_subprocess()
        sub.started.wait()
        self.assertEquals(True, sub.kill())
        sub.stopped.wait()
        self.assertEquals(-signal.SIGTERM, sub.retcode)

    def test_kill_SIGKILL(self):
        sub = self.run_sleeper_in_subprocess()
        sub.started.wait()
        self.assertEquals(True, sub.kill(signal.SIGKILL))
        sub.stopped.wait()
        self.assertEquals(-signal.SIGKILL, sub.retcode)

    def test_kill_catch_zombie(self):
        kill_me = threading.Event()
        resume = threading.Event()
        class TestMonitor(icommon.SubprocessMonitor):

            def switch_and_set(self, new):
                super(TestMonitor, self).switch_and_set(new)
                if new is self.stopped:
                    # The process is running but we haven't called
                    # proc.communicate yet.
                    kill_me.set()
                    resume.wait()

        sub = self.run_sleeper_in_subprocess(klass=TestMonitor)
        self.addCleanup(sub.join, 0)
        sub.started.wait()
        kill_me.wait()
        # Kill the subprocess ourselves
        os.kill(sub.proc_pid, signal.SIGTERM)
        pid, st = os.waitpid(sub.proc_pid, 0)
        self.assertEquals(sub.proc_pid, pid)
        # Only the low byte contains the signal and the high bit if for core
        # dumps.
        self.assertEquals(signal.SIGTERM, st & 0x7F)
        # Release the thread so it can call proc.communicate
        resume.set()
        # Now pretend we don't know the subprocess is dead
        self.assertEquals(False, sub.kill())
        sub.join()
        # No useful retcode available
        self.assertEquals(None, sub.retcode)
        # Nothing either in stdout/stderr. This means that subprocesses should
        # handled SIGTERM to allow us to collect their data.
        self.assertIs(None, sub.out)
        self.assertIs(None, sub.err)


class TestSubprocessMonitorWithScript(tests.TestCaseInTempDir):

    def run_in_subprocess(self, args, klass=None):
        if klass is None:
            klass = icommon.SubprocessMonitor
        sub = klass(args)
        self.addCleanup(sub.join, 0)
        sub.start()
        return sub

    def test_kill_caught(self):
        self.build_tree_contents([('test.py',
                                   """\
class Killed(Exception):

    def __str__(self):

        return "I'm so dead"

def die(signum, frame):
    raise Killed
import signal
signal.signal(signal.SIGTERM, die)
f = open('ready_to_die', 'w')
try:
    f.write('Yes I am\\n')
finally:
    f.close()
import time
slice = 0.001 # seconds
max = 120.0 # seconds
# We want to die long after the test fail (it it succeeds we're dead already)
while xrange(max/slice):
    time.sleep(0.001)
""")])
        sub = self.run_in_subprocess(['test.py'])
        sub.started.wait()
        # Gross hack to ensure the script ran long enough to catch the signal
        ready = False
        while not ready:
            try:
                f = open('ready_to_die')
                ready = True
            except:
                pass
        self.assertEquals(True, sub.kill())
        sub.stopped.wait()
        # The signal has been caught, we just get a failure return code, not
        # funky stuff with the signal in the low byte and so on
        self.assertEquals(1, sub.retcode)
        # after the split, the last line is empty
        self.assertEndsWith(sub.err.split('\n')[-2], "I'm so dead")

class Sleeper(icommon.SubprocessMonitor):

    def __init__(self, sleep_time=0.001, max_duration=None):
        # sleep can't be killed (or interrupted) so we don't sleep for long but
        # keep sleeping until killed
        super(Sleeper, self).__init__(
            # Make sure we die long after the test finish (to avoid leaks) but
            # still use a value that allow *some* debug (you want more for
            # *heavy* debug :)
            ['-c', 'import time; while xrange(120.0/%s): time.sleep(%s)'
             % (sleep_time, sleep_time)],
            max_duration=max_duration)


class TestThreadDriver(tests.TestCase):

    def get_driver(self, *args, **kwargs):
        klass = kwargs.pop('klass', None)
        if klass is None:
            klass = icommon.ThreadDriver
        return klass(*args, **kwargs)

    def get_started_driver(self, target=None, override=None, queue=None,
                           max_threads=None, klass=None):
        if queue is None:
            # Creating a queue here roughly means nobody will ever touch it, so
            # it will stay empty. This allows tests about the behaviour of the
            # driver when no sub-threads are involved.
            queue = Queue.Queue(1)
        if max_threads is None:
            # Tests that don't specify max_threads don't expect threads at all
            max_threads = 0
        driver = self.get_driver(queue, max_threads, target=target, klass=klass)
        if override is not None:
            for meth_name, callback in override:
                setattr(driver, meth_name, callback)
        self.addCleanup(driver.join, 0)
        driver.start()
        return driver

    def test_start_stop(self):
        """A do-nothing driver can be started and stopped."""
        driver = self.get_started_driver()
        driver.started.wait()
        driver.stop.set()

    def test_broken_target(self):
        """Exception while starting is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(target=raise_my_exception)
        driver.started.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_to_start(self):
        """Exception in the starting hook is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(
            override=[('before_start', raise_my_exception)])
        driver.started.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_during_first_step(self):
        """Exception during the first single step is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(
            override=[('do_one_step', raise_my_exception)])
        driver.stopped.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_during_second_step(self):
        """Exception during the next steps is caught."""
        self.steps = 0
        def raise_my_exception():
            if self.steps > 1:
                raise ExceptionForTests()
            else:
                self.steps += 1
        driver = self.get_started_driver(
            override=[('do_one_step', raise_my_exception)])
        driver.stopped.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_fail_while_stopping(self):
        """Exception during shutdown is caught."""
        def raise_my_exception():
            raise ExceptionForTests()
        driver = self.get_started_driver(
            override=[('before_stop', raise_my_exception)])
        driver.stop.set()
        driver.stopped.wait()
        self.assertRaises(ExceptionForTests, driver.pending_exception)

    def test_one_thread_finish(self):
        self.started = threading.Event()
        self.resumed = threading.Event()
        self.terminated = threading.Event()
        class TestThread(icommon.SubprocessMonitor):

            def collect(thread):
                self.assertIs(self.thread, thread)
                self.assertEquals(1, thread.retcode)
                self.terminated.set()
        self.thread = TestThread(['-c', 'import sys; sys.exit(1)'])
        self.addCleanup(self.thread.join, 0)
        q = Queue.Queue(1)
        q.put(self.thread)

        class TestDriver(icommon.ThreadDriver):

            def start_new_thread(driver):
                t = super(TestDriver, driver).start_new_thread()
                # all calls to start_new_thread except one will hit an empty
                # queue
                if t is not None:
                    self.assertIs(self.thread, t)
                    self.started.set()
                    self.resumed.wait()

        driver = TestDriver(queue=q, max_threads=1)
        self.addCleanup(driver.join, 0)
        driver.start()
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        # Wait for the subprocess thread to start
        self.started.wait()
        # the thread is started and at most wait for collection on
        # self.terminated so it's still there
        self.assertLength(1, driver.threads)
        self.assertEquals(self.thread, driver.threads[0])
        self.resumed.set()
        self.terminated.wait()
        driver.stop.set()
        driver.stopped.wait()
        self.assertLength(0, driver.threads)

    def test_thread_exception_raised_in_driver(self):
        class TestThread(icommon.SubprocessMonitor):

            def collect(thread):
                self.assertEquals(1, thread.retcode)
                raise ExceptionForTests()
        q = Queue.Queue(1)
        t = TestThread(['-c', 'import sys; sys.exit(1)'])
        self.addCleanup(t.join, 0)
        q.put(t)
        driver = self.get_started_driver(queue=q, max_threads=1)
        t.stopped.wait()
        driver.stop.set()
        self.assertRaises(ExceptionForTests, driver.join)

    def test_kill_pending_threads(self):
        q = Queue.Queue(3)
        for i in range(3):
            s = Sleeper()
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill)
            q.put(s)
        driver = self.get_started_driver(queue=q, max_threads=3)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        # Wait for the driver to empty the queue
        q.join()
        self.assertLength(3, driver.threads)
        # Tell the driver to stop *now*: kill the threads
        driver.stop.set()
        driver.stopped.wait()
        self.assertLength(0, driver.threads)

    def test_respect_max_threads(self):
        q = Queue.Queue(3)
        driver = self.get_started_driver(queue=q, max_threads=2)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        self.assertLength(0, driver.threads)
        sleepers = [Sleeper() for _ in range(3)]
        for s in sleepers:
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill)
            q.put(s)
        sleepers[0].started.wait()
        sleepers[1].started.wait()
        self.assertLength(2, driver.threads)
        self.assertEquals([sleepers[0], sleepers[1]], driver.threads)
        # sleepers[2] is still... awake
        self.assertFalse(sleepers[2].started.isSet())
        sleepers[0].kill()
        sleepers[0].stopped.wait()
        # Now the last sleeper will be started
        sleepers[2].started.wait()
        self.assertLength(2, driver.threads)
        self.assertEquals([sleepers[1], sleepers[2]], driver.threads)

    def test_respect_lowering_max_threads(self):
        q = Queue.Queue(4)
        sleepers = [Sleeper() for _ in range(4)]
        for s in sleepers:
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill)
            q.put(s)
        driver = self.get_started_driver(queue=q, max_threads=3)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        sleepers[0].started.wait()
        sleepers[1].started.wait()
        sleepers[2].started.wait()
        self.assertLength(3, driver.threads)
        self.assertEquals([sleepers[0], sleepers[1], sleepers[2]],
                          driver.threads)
        driver.max_threads = 2
        # Kill 2 sleepers and wait
        sleepers[0].kill()
        sleepers[0].stopped.wait()
        sleepers[1].kill()
        sleepers[1].stopped.wait()
        # Wait for another sleeper to start (because it's easier to sync on a
        # sleeper start than trying to catch when a sleeper goes out of the
        # driver's scope)
        sleepers[3].started.wait()
        self.assertEquals([sleepers[2], sleepers[3]], driver.threads)

    def test_closing_queue_stop(self):
        q = Queue.Queue(3)
        sleepers = [Sleeper() for _ in range(3)]
        for s in sleepers[:-1]: # the third sleeper will never be run
            self.addCleanup(s.join, 0)
            self.addCleanup(s.kill)
            q.put(s)
        driver = self.get_started_driver(queue=q, max_threads=2)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        sleepers[0].started.wait()
        sleepers[1].started.wait()
        # close the queue
        driver.queue_closed.set()
        sleepers[0].kill()
        sleepers[1].kill()
        driver.stopped.wait()

    def test_check_quotas(self):
        # A sleeper that will be killed
        sleeper = Sleeper(sleep_time=1, max_duration=0.0)
        q = Queue.Queue(1)
        self.addCleanup(sleeper.join, 0)
        self.addCleanup(sleeper.kill)
        q.put(sleeper)
        self.killed = None
        class TestDriver(icommon.ThreadDriver):

            def check_time_quota(driver):
                killed = super(TestDriver, driver).check_time_quota()
                if killed:
                    self.killed = killed
        driver = self.get_started_driver(klass=TestDriver,
                                         queue=q, max_threads=2)
        driver.started.wait()
        self.addCleanup(driver.stop.set)
        self.assertLength(0, driver.threads)
        sleeper.started.wait()
        # wait longer than the allowed duration
        time.sleep(0.01)
        self.assertEquals([sleeper], self.killed)
        driver.stop.set()
        driver.stopped.wait()


# FIXME: The ImportController is a bit harder to test as it handles signals and
# as such requires being run in the main thread.
class TestController(tests.TestCaseInTempDir):

    pass


if __name__ == '__main__':
    unittest.TestProgram(module=__name__)


